#include <torch/extension.h>
#include <torch/torch.h>
#include <vector>

// This combined kernel pre-initializes all LSTM layers and then sequentially applies them.
// Pre-initialization avoids repeated instantiation inside the loop, reducing overhead.

torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    // Ensure initial hidden and cell states are on the same device as x
    h0 = h0.to(x.device());
    c0 = c0.to(x.device());

    auto out = x;
    auto hn = h0.clone();
    auto cn = c0.clone();

    const size_t num_layers = lstm_weights_ih.size();
    std::vector<torch::nn::LSTM> lstm_layers;
    lstm_layers.reserve(num_layers);

    // Pre-initialize all LSTM layers to avoid per-iteration instantiation overhead
    for (size_t i = 0; i < num_layers; ++i) {
        int64_t input_size = lstm_weights_ih[i].size(1);
        int64_t hidden_size = lstm_weights_hh[i].size(1);

        // Create LSTM layer with single layer options
        auto lstm_layer = torch::nn::LSTM(
            torch::nn::LSTMOptions(input_size, hidden_size)
                .num_layers(1)
                .batch_first(true)
                .bidirectional(false)
        );
        lstm_layer->to(x.device());

        // Copy provided weights and biases
        lstm_layer->named_parameters()["weight_ih_l0"].copy_(lstm_weights_ih[i]);
        lstm_layer->named_parameters()["weight_hh_l0"].copy_(lstm_weights_hh[i]);
        lstm_layer->named_parameters()["bias_ih_l0"].copy_(lstm_biases_ih[i]);
        lstm_layer->named_parameters()["bias_hh_l0"].copy_(lstm_biases_hh[i]);

        lstm_layer->train(is_training);
        lstm_layers.push_back(lstm_layer);
    }

    // Process each LSTM layer sequentially using pre-initialized modules
    for (size_t i = 0; i < num_layers; ++i) {
        // Slice hidden and cell states corresponding to layer i
        auto h_slice = hn.narrow(0, i, 1);
        auto c_slice = cn.narrow(0, i, 1);
        std::tuple<torch::Tensor, torch::Tensor> state_tuple = std::make_tuple(h_slice, c_slice);

        // Forward pass for current layer
        auto output_and_state = lstm_layers[i]->forward(out, state_tuple);
        out = std::get<0>(output_and_state);
        auto h_n = std::get<0>(std::get<1>(output_and_state));
        auto c_n = std::get<1>(std::get<1>(output_and_state));

        // Update hidden states in-place
        hn.narrow(0, i, 1).copy_(h_n);
        cn.narrow(0, i, 1).copy_(c_n);
    }

    return hn;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused LSTM forward (CUDA)");
}
